"""Data analysis Vildhede2015."""
import numpy as np
import pandas as pd

from protein_distribution import DATA_DIR
from protein_distribution.uniprot import read_uniprot_mapping


def parse_vildhede2015() -> None:
    """Parse raw data from supplementary files."""

    tsv_in_path = DATA_DIR / "Vildhede2015" / "raw" / "proteinGroups.tsv"
    xlsx_processed_path = DATA_DIR / "Vildhede2015" / "Vildhede2015_processed.xlsx"
    xlsx_out_path = DATA_DIR / "Vildhede2015" / "Vildhede2015_data.xlsx"

    if not tsv_in_path.exists():
        raise IOError(f"File does not exist: {tsv_in_path}")

    # read protein mapping
    uniprot2sid = read_uniprot_mapping()

    # read data and process
    df_raw = pd.read_csv(tsv_in_path, sep="\t")
    columns = [
        "Protein IDs",  # Uniprot protein identifiers: mapping with protein_overview.xlsx
        "Protein names",
        "Gene names",
        "Proteins",  # protein count
        "Mol. weight [kDa]",  # Mr [g/mole]  # 1 dalton = 1.66053000000133E-24 gram [g].
        "Intensity",
    ]
    experiments_hepatocyte = [
        # 8 hepatocyte membrance fractions (3 repeats 1, 2, 3)
        "HA1",
        "HA2",
        "HA3",
        "HB1",
        "HB2",
        "HB3",
        "HC1",
        "HC2",
        "HC3",
        "HD1",
        "HD2",
        "HD3",
        "HE1",
        "HE2",
        "HE3",
        "HF1",
        "HF2",
        "HF3",
        "HG1",
        "HG2",
        "HG3",
        "HH1",
        "HH2",
        "HH3",
    ]
    experiments_liver = [
        # 12 liver membrane fractions
        "L1",
        "L2",
        "L3",
        "L4",
        "L5",
        "L6",
        "L7",
        "L8",
        "L9",
        "L10",
        "L11",
        "L12",
    ]
    experiments = experiments_hepatocyte + experiments_liver
    columns = columns + [f"Intensity {exp}" for exp in experiments]
    print(columns)
    print("-" * 80)

    # get subset of columns
    df = df_raw[columns]
    print(df.head(10))
    print("-" * 80)

    # normalization to [mole/g (protein)]
    for exp in experiments:
        total_signal = df[f"Intensity {exp}"].sum()
        mr = df["Mol. weight [kDa]"] * 1000  # [Da] = [g/mole]

        # Calculation of protein amount from intensity data
        df[exp] = (
            (df[f"Intensity {exp}"]) / (total_signal * mr) * 1e9
        )  # [mole/g] => pmol/mg = nmol/g

    # calculate the hepatocyte mean values
    for k, key in enumerate(["A", "B", "C", "D", "E", "F", "G", "H"]):
        # calculate mean
        df[f"H{k+1}"] = df[f"H{key}1"] + df[f"H{key}2"] + df[f"H{key}3"] / 3

        # Filter by complete replicates
        zero_index = (
            (df[f"H{key}1"] == 0) | (df[f"H{key}2"] == 0) | (df[f"H{key}3"] == 0)
        )
        df.loc[zero_index, f"H{k+1}"] = 0

    # delete old keys
    df.drop([f"Intensity {key}" for key in experiments], axis=1, inplace=True)
    df.drop([f"{key}" for key in experiments_hepatocyte], axis=1, inplace=True)
    print(df.head(10))
    print("-" * 80)

    # get subset of proteins (filter protein of interests
    sids = []
    for _, row in df.iterrows():
        # FIXME: check if in place
        proteins = row["Protein IDs"]
        protein_ids = set(proteins.split(";"))
        for uniprot, sid in uniprot2sid.items():
            if uniprot in protein_ids:
                sids.append(sid)
                break
        else:
            sids.append(np.NaN)

    # print(sids)
    df["sid"] = sids

    # filter dataframe
    df_filtered = df[~pd.isna(df.sid)]

    # store results
    with pd.ExcelWriter(xlsx_processed_path) as writer:
        df_filtered.to_excel(writer, sheet_name="protein_amounts", index=False)

    # store results in our format

    data = {
        "study": "Vildhede2015",
        "group": None,
        "individual": None,  # fill this,
        "count": 1,
        "measurement_type": "abundance",
        "protein": None,  # fill this
        "tissue": None,  # fill this (liver microsomes, jejunun),
        "method": None,
        "value": None,  # fill this
        "mean": None,
        "sd": None,
        "median": None,
        "min": None,
        "max": None,
        "mad": None,
        "cv": None,
        "cv_unit": None,
        "unit": "pmol/mg",
        "comments": None,
    }
    info_items = []

    exp_columns = [
        "H1",
        "H2",
        "H3",
        "H4",
        "H5",
        "H6",
        "H7",
        "H8",
        "L1",
        "L2",
        "L3",
        "L4",
        "L5",
        "L6",
        "L7",
        "L8",
        "L9",
        "L10",
        "L11",
        "L12",
    ]
    for _, row in df_filtered.iterrows():
        sid = row["sid"]
        for exp in exp_columns:
            value = row[exp]
            if np.isclose(value, 0.0):
                continue
            d = {**data}
            d["individual"] = exp
            d["protein"] = sid
            d["tissue"] = "liver microsomes" if exp.startswith("L") else "hepatocyte"
            d["value"] = value
            info_items.append(d)

    df_protein = pd.DataFrame(info_items)
    with pd.ExcelWriter(xlsx_out_path) as writer:
        df_protein.to_excel(writer, sheet_name="protein_amount", index=False)


if __name__ == "__main__":
    parse_vildhede2015()



"""pymetadata - Python utilities for metadata."""
from pathlib import Path

__author__ = "Matthias Koenig"
__version__ = "0.1.2"


program_name: str = "protein-distribution"

BASE_DIR = Path(__file__).parent.parent.parent

# data for analysis (not distributed)
DATA_DIR = BASE_DIR / "data"
PROTEIN_PATH = DATA_DIR / "protein_overview.xlsx"

# results of analysis
RESULTS_DIR = Path(__file__).parent.parent.parent / "results"
RESULTS_DIR.mkdir(exist_ok=True)

# resources (distributed)
RESOURCES_DIR = Path(__file__).parent / "resources"

UNIPROT_PATH = RESOURCES_DIR / "uniprot.json"
UNIPROT_MAPPING_PATH = RESOURCES_DIR / "uniprot_mapping.json"

DATA_RAW_XLSX = RESOURCES_DIR / "data_raw.xlsx"
DATA_XLSX = RESOURCES_DIR / "data.xlsx"
DATA_MERGED_XLSX = RESOURCES_DIR / "data_merged.xlsx"
DATA_NORMALIZATION_XLSX = RESOURCES_DIR / "data_normalization.xlsx"



"""Parse data from supplentary file of Wegler2022."""
from pathlib import Path

import numpy as np
import pandas as pd

from protein_distribution import DATA_DIR


def parse_data() -> None:
    """Parse Wegler2022 proteomics data."""
    xlsx_in_path = DATA_DIR / "Wegler2022" / "Wegler2022_supplement_2.xlsx"
    xlsx_out_path = DATA_DIR / "Wegler2022" / "Wegler2022_data.xlsx"

    if not xlsx_in_path.exists():
        raise IOError(f"File does not exist: {xlsx_in_path}")

    dfs = {}
    subject_prefixes = ["Jejunum", "Liver", "CholLiver"]
    for k, sheet_name in enumerate(
        [
            "Proteins_Jejunum_Obese",
            "Proteins_Liver_Obese",
            "Proteins_Liver_NonObese",
        ]
    ):
        subject_prefix = subject_prefixes[k]
        print(subject_prefix)

        df = pd.read_excel(xlsx_in_path, sheet_name=sheet_name, skiprows=[0])
        print(df.head())

        column_subset = {}
        for column in df.columns:
            if column == "Gene.names":
                column_subset[column] = "gene"

            elif column.startswith(subject_prefix):
                tokens = column.split(" ")
                k = tokens[1]
                if sheet_name in ["Proteins_Jejunum_Obese", "Proteins_Liver_Obese"]:
                    column_subset[column] = f"Obese{k}"
                elif sheet_name in ["Proteins_Liver_NonObese"]:
                    column_subset[column] = f"Nonobese{k}"

        # create subset
        df_filtered = df[list(column_subset.keys())]

        # rename columns
        df_filtered.rename(columns=column_subset, inplace=True)
        # set index
        df_filtered.set_index(keys=["gene"], inplace=True)

        print(df_filtered.head())

        # TabA, TabB, TabC
        data = {
            "study": "Wegler2022",
            "group": None,
            "individual": None,  # fill this,
            "count": 1,
            "measurement_type": "abundance",
            "protein": None,  # fill this
            "tissue": None,  # fill this (liver microsomes, jejunun),
            "method": None,
            "value": None,  # fill this
            "mean": None,
            "sd": None,
            "median": None,
            "min": None,
            "max": None,
            "mad": None,
            "cv": None,
            "cv_unit": None,
            "unit": "fmol/µg",
            "comments": None,
        }
        info_items = []

        for gene, row in df_filtered.iterrows():
            # FIXME: this must be done via the protein idenifiers
            if (
                gene.startswith("CYP")
                or gene.startswith("UGT")
                or gene.startswith("ABC")
                or gene.startswith("SLC")
            ):
                if ";" in gene:
                    # skip combined proteins
                    continue
                for column in df_filtered.columns:
                    value = row[column]
                    if np.isnan(value):
                        # skip empty values
                        continue
                    d = {**data}
                    d["individual"] = column
                    d["protein"] = gene
                    d["tissue"] = (
                        "jejunum" if "Jejunum" in sheet_name else "liver microsomes"
                    )
                    d["value"] = row[column]
                    info_items.append(d)

        df_protein = pd.DataFrame(info_items)
        dfs[sheet_name] = df_protein
        print(df_protein.to_string())

    tabs = ["TabA", "TabB", "TabC"]
    with pd.ExcelWriter(xlsx_out_path) as writer:
        for k, sheet_name in enumerate(dfs):
            df = dfs[sheet_name]
            df.to_excel(writer, sheet_name=tabs[k], index=False)


if __name__ == "__main__":
    parse_data()



"""Module for retrieving and working with protein information."""

from typing import Dict, List

import pandas as pd

from protein_distribution import PROTEIN_PATH, log
from protein_distribution.console import console
from protein_distribution.uniprot import read_uniprot_mapping


logger = log.get_logger(__name__)


def get_proteins_uniprot(proteins: List[str]) -> List[str]:
    """Filter a list of protein identifiers.

    Only returns identifiers with uniprot entry.
    """
    # get subset of proteins and filter for entries with uniprot id
    uniprot2sid = read_uniprot_mapping()
    sid2uniprot = {v: k for k, v in uniprot2sid.items()}

    proteins_uniprot = [p for p in proteins if p in sid2uniprot]
    # proteins_filtered = [p for p in proteins if p not in sid2uniprot]
    # console.print(f"Filtered UniProt: {proteins_filtered}")

    return proteins_uniprot


def get_proteins(df_abundance: pd.DataFrame, uniprot: bool = True) -> List[str]:
    """Get unique list of protein identifiers from Abundance DataFrame."""
    proteins = sorted(df_abundance["protein"].unique())
    if uniprot:
        proteins = get_proteins_uniprot(proteins)
    return proteins


def get_protein_categories(proteins: List[str]) -> Dict[str, List[str]]:
    """Categorizes proteins in classes.

    Expects proteins with uniprot identifier.
    """
    df_proteins = pd.read_excel(
        PROTEIN_PATH, sheet_name="proteins", skiprows=[0], comment="#"
    )

    sid2category = dict(zip(df_proteins["name"], df_proteins["category"]))
    d: Dict[str, List[str]] = {}
    for sid in proteins:
        category = sid2category[sid]
        if category in d:
            d[category].append(sid)
        else:
            d[category] = [sid]

    return d



"""Uniprot information."""
import json
from typing import Dict, List

import numpy as np
import pandas as pd
import requests
import xmltodict

# from dataclasses import dataclass
from pydantic import BaseModel

from protein_distribution import PROTEIN_PATH, UNIPROT_MAPPING_PATH, UNIPROT_PATH
from protein_distribution.console import console


def create_uniprot_mapping() -> Dict[str, str]:
    """Read protein mapping.

    For some CYPs no uniprot information was available. These were:
    - CYP1A6
    - CYP4F
    - CYP2C
    """

    df_proteins = pd.read_excel(
        PROTEIN_PATH,
        sheet_name="proteins",
        skiprows=[0],
        comment="#",
        # dtype={"uniprot": "string", "short": "string", "fullname": "string"}
    )
    df_proteins.replace("-", "", inplace=True)
    uniprot2sid: Dict[str, str] = {}
    for uniprot, name in zip(df_proteins["uniprot"], df_proteins["name"]):
        if uniprot and name and uniprot != "NaN":
            uniprot2sid[uniprot] = name

    with open(UNIPROT_MAPPING_PATH, "w") as f_json:
        json.dump(uniprot2sid, f_json, indent=2)

    return uniprot2sid


def read_uniprot_mapping() -> Dict[str, str]:
    """Read the uniprot mapping."""
    with open(UNIPROT_MAPPING_PATH, "r") as f_json:
        d: Dict[str, str] = json.load(f_json)
    return d


class UniprotMetadata(BaseModel):
    """Metadata for Uniprot Entry."""

    id: str  # P05177
    name: str  # CP1A2_HUMAN
    protein: str  # Cytochrome P450 1A2
    gene: str  # CYP1A2
    organism: str  # Homo sapiens (Human)
    # evidence: str
    function: str


class UniprotMetadataDict(BaseModel):
    """Dictionary of Uniprot Metadata entries."""

    entries: Dict[str, UniprotMetadata]


def query_uniprot_metadata(uniprot_id: str) -> UniprotMetadata:
    """Parse Metadata for protein identifier."""
    url: str = f"https://rest.uniprot.org/uniprotkb/{uniprot_id}.xml"
    console.log(f"Query: {url}")

    response = requests.get(url)
    response.raise_for_status()
    d = xmltodict.parse(response.content)

    try:
        entry = d["uniprot"]["entry"]
        organism_tokens = entry["organism"]["name"]
        protein = entry["protein"]
        if "recommendedName" in protein:
            fullname = entry["protein"]["recommendedName"]["fullName"]
        elif "submittedName" in protein:
            fullname = entry["protein"]["submittedName"]["fullName"]
        if isinstance(fullname, dict):
            fullname = fullname["#text"]

        genename = entry["gene"]["name"]
        if isinstance(genename, str):
            gene = genename
        elif isinstance(genename, list):
            gene = ", ".join(g["#text"] for g in genename)
        elif isinstance(genename, dict):
            gene = genename["#text"]

        comments = entry["comment"]
        function = ""
        if isinstance(comments, dict):
            if comments["@type"] == "function":
                comment = comments["text"]
                if isinstance(comment, str):
                    function = comment
                elif isinstance(comment, dict):
                    function = comment["#text"]
        elif isinstance(comments, list):
            for c in comments:
                if c["@type"] == "function":
                    comment = c["text"]
                    if isinstance(comment, str):
                        function = comment
                    elif isinstance(comment, dict):
                        function = comment["#text"]
                    break

        md = UniprotMetadata(
            id=uniprot_id,
            name=entry["name"],
            protein=fullname if isinstance(fullname, str) else fullname["#text"],
            gene=gene,
            organism=f"{organism_tokens[0]['#text']} ({organism_tokens[1]['#text']})",
            function=function,
        )
        # console.print(md)
    except Exception as err:
        console.print(entry)
        raise err

    return md


def query_metadata() -> UniprotMetadataDict:
    """Query Uniprot metadata via webservices."""
    uniprot2sid = read_uniprot_mapping()

    mds = {}
    for uniprot_id in uniprot2sid.keys():
        mds[uniprot_id] = query_uniprot_metadata(uniprot_id)

    mddict = UniprotMetadataDict(entries=mds)
    with open(UNIPROT_PATH, "w") as f_json:
        f_json.write(mddict.model_dump_json(indent=2))

    return mddict


def read_metadata() -> UniprotMetadataDict:
    """Read uniprot metadata."""

    with open(UNIPROT_PATH, "r") as f_json:
        d = json.load(f_json)

    return UniprotMetadataDict(**d)


if __name__ == "__main__":
    # mapping
    create_uniprot_mapping()
    mapping = read_uniprot_mapping()

    # uniprot metadata
    query_metadata()
    mddict = read_metadata()
    console.print(mddict)



"""Example visualization of data.

TODO: size of mean groups & annotations for mean group size
TODO: separate mean and individual data
TODO: calculate mean from individual data
TODO: fitting of histogram (lognormal)
"""
from pathlib import Path
from typing import Iterable, Optional

# ------------------------------------------------------------------------------
# plt.style.use('science')
import matplotlib
import matplotlib.axes
import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from protein_distribution.console import console
from protein_distribution.protein_info import get_protein_categories, get_proteins
from protein_distribution.significance_testing import (
    process_stratification_data,
    stratification_info,
)


SMALL_SIZE = 12
MEDIUM_SIZE = 15
BIGGER_SIZE = 25

matplotlib.rc("font", size=SMALL_SIZE)  # controls default text sizes
matplotlib.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
matplotlib.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
# ------------------------------------------------------------------------------


# handle colors
cmap = matplotlib.cm.get_cmap("tab20")
studies = [
    "Achour2014",
    "Achour2017",
    "Asai1996",
    "Couto2019",
    "Fallon2016",
    "Gao2016",
    "Groer2014",
    "Inoue2000",
    "Kawakami2011",
    "Karlgren2012",
    "Kawakami2011",
    "Kimoto2012",
    "Li2015",
    "Michaels2014",
    "Ohtsuki2012",
    "Olesen2000",
    "Prasad2014",
    "Seibert2009",
    "Snawder2000",
    "Takahashi2021",
    "Tucker2012",
    "Vasilogianni2022",
    "Vildhede2015",
    "Wang2015",
    "Wegler2022",
]
study2color = {study: cmap(k / 20) for k, study in enumerate(studies)}


def plot_protein_abundance(
    data: pd.DataFrame, protein_id: str, image_dir: Optional[Path]
) -> Figure:
    """Plot data for a single enzyme for multiple studies."""

    # subset of abundance for protein
    df_protein = data[data.protein == protein_id]
    sids = list(reversed(sorted(df_protein.sid.unique())))

    # figsize=(10, 0.4 * len(sids))
    height = max(0.4 * len(sids), 3)
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, height), dpi=300)
    fig.subplots_adjust(left=0.3, bottom=0.2)
    labels = []
    for k, sid in enumerate(sids):
        study = sid.split("_")[0]
        color = study2color[study]
        df_sid = df_protein[df_protein.sid == sid]
        y = k * np.ones(shape=(len(df_sid)))

        values = df_sid["value"].values
        means = df_sid["mean"].values
        sds = df_sid["sd"].values

        # plot individual data
        kwargs = {
            "linestyle": "None",
            "markeredgecolor": "black",
            "markersize": 9,
            "color": color,
        }

        if not np.all(np.isnan(values)):
            ax.plot(values, y, marker="o", alpha=0.7, **kwargs)
            # annotate points
            # for i in range(len(y)):
            #    ax.annotate(sid, (x[i], y_values[i]))

        # plot mean data
        # FIXME: add SD (calculate SD in dataset)
        # FIXME: add [min-max] if available
        if not np.all(np.isnan(means)):
            # ax.plot(means, y, marker="s", alpha=1.0, **kwargs)
            ax.errorbar(
                x=means, y=y, xerr=sds, marker="s", alpha=0.8, ecolor="black", **kwargs
            )
            # for i in range(len(x)):
            #     ax.annotate(sid, (x[i], y_means[i]))
        n_values = np.count_nonzero(~np.isnan(values))
        n_means = np.count_nonzero(~np.isnan(means))
        labels.append(
            f"{sid}\nn={n_means if n_means else ' '}|{n_values if n_values else ' '}"
        )

    # ax.set_xlim(left=0)
    ax.set_xlabel("Protein abundance [pmol/mg]", fontdict={"weight": "bold"})
    # ax.set_ylabel("Dataset", fontdict={"weight": "bold"})
    ax.set_title(protein_id, fontdict={"weight": "bold", "size": 20})
    # ax.legend(loc=(1.04, 0))
    ax.set_yticks(
        ticks=range(len(labels)),
        labels=labels,
        weight="bold",
        size=8,
    )

    # plt.show()
    if image_dir:
        fig.savefig(image_dir / f"{protein_id}.svg")
        fig.savefig(image_dir / f"{protein_id}.png", bbox_inches="tight")
    return fig


def plot_protein_histogram(
    data: pd.DataFrame, protein_id: str, image_dir: Optional[Path]
) -> Optional[Figure]:
    """Plot histogram."""

    # subset of abundance for individual protein
    df_protein = data[(data.protein == protein_id) & (~pd.isnull(data.value))]
    studies = list(sorted(df_protein.study.unique()))
    console.print(studies)

    # get all values
    values_list = []
    labels = []
    colors = []
    # FIXME: only use individual values
    for _, study in enumerate(studies):
        # study = sid.split("_")[0]
        colors.append(study2color[study])
        df_sid = df_protein[df_protein.study == study]

        values = df_sid["value"].values
        print(study, values)
        values_list.append(values)
        labels.append(study)

    if not values_list:
        console.log(f"No data for '{protein_id}', no plot generated.")
        return None

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 7), dpi=300)
    # fig.subplots_adjust(left=0.2, bottom=0.2)

    ax.hist(
        values_list,
        bins=20,
        stacked=True,
        density=True,
        color=colors,
        edgecolor="black",
        label=labels,
    )

    ax.set_xlabel("Protein abundance [pmol/mg]", fontdict={"weight": "bold"})
    ax.set_ylabel("Density", fontdict={"weight": "bold"})
    ax.set_title(protein_id, fontdict={"weight": "bold", "size": 20})
    ax.legend()
    if image_dir:
        fig.savefig(image_dir / f"{protein_id}_hist.svg")
        fig.savefig(image_dir / f"{protein_id}_hist.png", bbox_inches="tight")
    return fig


def plot_correlation(
    data: pd.DataFrame, pid1: str, pid2: str, image_dir: Optional[Path]
) -> Figure:
    """Plot correlation."""

    df = data[["study", "individual", "tissue", pid1, pid2]]
    # only individual data
    row_index = ~pd.isnull(df[pid1]) & ~pd.isnull(df[pid2])
    df = df[row_index]

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 5), dpi=300)
    fig.subplots_adjust(left=0.2, bottom=0.2)
    for study in df.study.unique():
        df_subset = df[df.study == study]
        ax.plot(
            df_subset[pid1],
            df_subset[pid2],
            "o",
            markeredgecolor="black",
            label=study,
            alpha=0.8,
            markersize=10,
        )
    ax.set_xlabel(f"{pid1} [pmol/mg]", fontdict={"weight": "bold"})
    ax.set_ylabel(f"{pid2} [pmol/mg]", fontdict={"weight": "bold"})
    ax.legend(prop={"size": 8})
    plt.show()
    if image_dir:
        fig.savefig(image_dir / f"correlation_{pid1}_{pid2}.svg")
        fig.savefig(image_dir / f"correlation_{pid1}_{pid2}.png", bbox_inches="tight")


def plot_multi_stratified(
    data: pd.DataFrame,
    category: str,
    image_dir: Optional[Path],
    stratification_key: str = "sex",
) -> Figure:
    """Plot multiple proteins stratified by information.

    existing groups and filtered groups:

    sex: [M, F, NR] -> [M, F]
    smoking [Y, N, NR] -> [Y, N]
    alcohol [Y, N, NR] -> [Y, N]
    ethnicity [caucasian, hispanic, african american, NR] -> [caucasian, african american, hispanic]
    age_group [adolescent, young, middle aged, elderly, NR] -> [middle aged, elderly]
    bmi_group [underweight, normal weight, overweight, obese, NR] -> [normal weight, overweight, obese]
    """

    # process stratification data
    df_protein = process_stratification_data(
        data=data, stratification_key=stratification_key
    )

    # order by mean
    means = {}
    for p in df_protein.protein.unique():
        means[p] = df_protein[df_protein.protein == p].value.mean()
    keys = list(means.keys())
    values = list(means.values())
    sorted_value_index = np.argsort(values)
    proteins_sorted = list(reversed([keys[i] for i in sorted_value_index]))

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(20, 10), dpi=300)
    fig.subplots_adjust(left=0.2, bottom=0.2)

    # creating boxplot
    groups = stratification_info[stratification_key]["groups"]
    colors = stratification_info[stratification_key]["colors"]
    sns.boxplot(
        ax=ax,
        y="value",
        x="protein",
        hue=stratification_key,
        hue_order=groups,
        data=df_protein,
        order=proteins_sorted,
        palette=colors,
    )

    # adding data points
    sns.stripplot(
        ax=ax,
        y="value",
        x="protein",
        hue=stratification_key,
        hue_order=groups,
        dodge=True,
        data=df_protein,
        order=proteins_sorted,
        color="black",
        edgecolor="black",
        size=5,
        alpha=0.5,
    )

    ax.set_xlabel("Protein abundance [pmol/mg]", fontdict={"weight": "bold"})
    ax.set_ylabel("Protein", fontdict={"weight": "bold"})

    n_proteins = len(df_protein.protein.unique())
    title = f"{category} | {stratification_key} (m={n_proteins})"
    ax.set_title(title, fontdict={"weight": "bold", "size": 20})
    ax.set_yscale("log")
    ax.set_xticklabels(ax.get_xticklabels(), rotation=90, horizontalalignment="center")
    plt.show()

    if image_dir:
        fig.savefig(image_dir / f"{category}_boxplot_{stratification_key}.svg")
        fig.savefig(
            image_dir / f"{category}_boxplot_{stratification_key}.png",
            bbox_inches="tight",
        )

    return fig


def plot_multi(
    data: pd.DataFrame,
    title: str,
    image_dir: Optional[Path],
) -> Figure:
    """Plot multiple proteins as boxplot."""

    # subset of abundance for individual protein
    df_protein = data[~pd.isnull(data.value)]

    # order by mean
    means = {}
    for p in df_protein.protein.unique():
        means[p] = df_protein[df_protein.protein == p].value.mean()
    keys = list(means.keys())
    values = list(means.values())
    sorted_value_index = np.argsort(values)
    proteins_sorted = list(reversed([keys[i] for i in sorted_value_index]))

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(10, 20), dpi=300)
    fig.subplots_adjust(left=0.2, bottom=0.2)

    # creating boxplot
    sns.boxplot(ax=ax, x="value", y="protein", data=df_protein, order=proteins_sorted)

    # adding data points
    sns.stripplot(
        ax=ax,
        x="value",
        y="protein",
        data=df_protein,
        order=proteins_sorted,
        # jitter=2.0,
        color="black",
        edgecolor="black",
        size=5,
        alpha=0.5,
    )

    ax.set_xlabel("Protein abundance [pmol/mg]", fontdict={"weight": "bold"})
    ax.set_ylabel("Protein", fontdict={"weight": "bold"})
    ax.set_title(title, fontdict={"weight": "bold", "size": 20})
    ax.set_xscale("log")
    plt.show()

    category = title.split(" ")[0]
    if image_dir:
        fig.savefig(image_dir / f"{category}_boxplot.svg")
        fig.savefig(image_dir / f"{category}_boxplot.png", bbox_inches="tight")

    return fig


def run_all_visualization() -> None:
    """Run all visualizations."""
    # Load data
    from protein_distribution import DATA_MERGED_XLSX, DATA_XLSX, RESULTS_DIR

    df_abundance = pd.read_excel(DATA_XLSX, sheet_name="Abundance")
    del df_abundance["comments"]
    df_abundance["sid"] = df_abundance["study"] + "_" + df_abundance["source"]

    # data_group = pd.read_excel(DATA_MERGED_XLSX, sheet_name="group_data")
    # data_individual = pd.read_excel(DATA_MERGED_XLSX, sheet_name="group_data")

    proteins = get_proteins(df_abundance, uniprot=True)
    protein_categories = get_protein_categories(proteins)

    # Create plots
    # for protein_id in ["CYP1A2", "CYP2D6", "CYP2E1"]:
    image_dir = RESULTS_DIR / "images"
    image_dir.mkdir(parents=True, exist_ok=True)

    # plot_protein_abundance(
    #     data=df_abundance,
    #     protein_id="CYP2E1",
    #     image_dir=image_dir,
    # )
    # plot_protein_histogram(
    #     data=df_abundance,
    #     protein_id="CYP2E1",
    #     image_dir=image_dir,
    # )

    # two protein correlation plots
    if True:
        # with merged data
        correlation_individuals: pd.DataFrame = pd.read_excel(
            DATA_MERGED_XLSX, sheet_name="individual_correlation"
        )
        for pids in [
            # example cyps
            ["CYP2C8", "CYP1A2"],  # positive
            ["CYP4A11", "CYP2C8"],  # positive
            ["CYP3A4", "CYP2B6"],  # positive
            ["CYP2A7", "CYP27A1"],  # negative
            ["CYP2U1", "CYP2D6"],  # negative
            # example ugts
            ["UGT2B15", "UGT1A9"],  # positive
            ["UGT2B7", "UGT2B4"],  # positive
            ["UGT1A9", "UGT1A1"],  # positive
            ["UGT2A3", "UGT1A4"],  # negative
            # example abcs
            ["ABCA8", "ABCB11"],
            ["ABCB7", "ABCE1"],
            # example slcs
            ["SLC27A4", "SLCO1B3"],
        ]:
            plot_correlation(
                data=correlation_individuals,
                pid1=pids[0],
                pid2=pids[1],
                image_dir=image_dir,
            )

    if True:
        # overview plots of categories
        # FIXME: update with merged data
        df_individual_data = pd.read_excel(
            DATA_MERGED_XLSX, sheet_name="individual_data"
        )
        for category in protein_categories.keys():
            # filter by category
            df_category = df_individual_data[
                df_individual_data.protein.isin(protein_categories[category])
            ]
            plot_multi(
                df_category,
                title=f"{category} (m={len(protein_categories[category])})",
                image_dir=image_dir,
            )
            plt.show()

    if True:
        df_individual_data = pd.read_excel(
            DATA_MERGED_XLSX, sheet_name="individual_data"
        )
        for category in ["cyp", "ugt", "slc", "abc"]:
            df_category = df_individual_data[
                df_individual_data.protein.isin(protein_categories[category])
            ]

            for stratification_key in [
                "sex",
                "smoking",
                "alcohol",
                "ethnicity",
                "age_group",
                "bmi_group",
            ]:
                plot_multi_stratified(
                    df_category,
                    stratification_key=stratification_key,
                    category=category,
                    image_dir=image_dir,
                )
                plt.show()

    if True:
        # FIXME: without merged data
        # individual plots
        for category, proteins in protein_categories.items():
            category_dir = image_dir / category
            category_dir.mkdir(parents=True, exist_ok=True)

            for protein_id in proteins:
                console.rule(protein_id, align="left", style="white")

                plot_protein_abundance(
                    data=df_abundance,
                    protein_id=protein_id,
                    image_dir=category_dir,
                )
                plot_protein_histogram(
                    data=df_abundance,
                    protein_id=protein_id,
                    image_dir=category_dir,
                )


if __name__ == "__main__":
    run_all_visualization()



"""Prepare protein abundance data for analysis.

- data reading
- data validation
- data integration
- data normalization (e.g. unit normalization)
"""
from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Tuple

import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from protein_distribution import DATA_DIR, log, protein_info
from protein_distribution.console import console


logger = log.get_logger(__name__)


def load_study_overview() -> pd.DataFrame:
    """Load the study overview data in table."""
    df_studies = pd.read_excel(DATA_DIR / "study_overview.xlsx", sheet_name=0)
    return df_studies


def get_studies_by_status(df: pd.DataFrame, status: List[str]) -> List[str]:
    """Return list with ids of studies."""
    df_filtered = df[df.status.isin(status)]

    studies: List[str] = df_filtered.study.values
    return studies


def load_data(
    study_ids: Iterable[str], data_dir: Optional[Path] = None
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """Load data for given studies."""
    if data_dir is None:
        data_dir = DATA_DIR

    study_ids_set = set(study_ids)

    for file in Path(data_dir).glob("*"):
        if file.is_dir():
            study_dir = file.name
            if study_dir not in study_ids_set:
                logger.warning(
                    f"Study directory does not exist in study_overview table '{study_dir}'."
                )

    data = {}
    for study_id in study_ids:
        console.rule(title=study_id, align="left", style="white")
        dfs: Dict[str, pd.DataFrame] = load_data_from_study(
            study_id=study_id, data_dir=data_dir
        )
        data[study_id] = dfs

    return data


def load_data_from_study(
    study_id: str, data_dir: Optional[Path] = None
) -> Dict[str, pd.DataFrame]:
    """Load data for given study.

    :param study_id: id of study
    :param data_dir: directory in which the data is located.
    """
    if data_dir is None:
        data_dir = DATA_DIR

    error = False

    # check that folder exists
    study_dir: Path = data_dir / study_id
    if not study_dir.exists():
        logger.error(f"Study directory does not exist for '{study_id}': '{study_dir}'")
        error = True
    if not study_dir.is_dir():
        logger.error(f"Study directory is not a directory: '{study_dir}'")
        error = True

    # check that PDF exist
    pdf_path = study_dir / f"{study_id}.pdf"
    if not pdf_path.exists():
        logger.error(f"PDF does not exist for '{study_id}': '{pdf_path}'")
        error = True

    # check that Excel exist
    xlsx_path = study_dir / f"{study_id}.xlsx"
    if not xlsx_path.exists():
        logger.error(f"XLSX does not exist for '{study_id}': '{xlsx_path}'")
        error = True

    # check that names are correct in folder
    for file in Path(study_dir).glob("*"):
        filename = file.name
        # don't check temporary files
        if filename.startswith(".~"):
            continue
        elif filename.startswith("~$"):
            continue

        if file.is_file():
            if not filename.startswith(f"{study_id}"):
                logger.error(f"File does not start with '{study_id}': '{file}'")
                error = True

    if error:
        console.print(f"[error]Errors in files, no data is loaded from '{study_id}'.")
        return {}
    else:
        console.print(f"[success]Files are correct for '{study_id}'.")

    # load all sheets from excel
    dfs: Dict[str, pd.DataFrame] = pd.read_excel(
        xlsx_path, sheet_name=None, comment="#"
    )

    # remove empty rows
    for df in dfs.values():
        df.dropna(how="all", axis="index", inplace=True)

    # check sheet names:
    sheet_names = dfs.keys()
    for sheet_name in sheet_names:
        if sheet_name in {"Groups", "Individuals"}:
            continue
        elif sheet_name.startswith("Tab"):
            continue
        elif sheet_name.startswith("Fig"):
            continue
        else:
            logger.error(
                f"Incorrect sheet_name in '{study_id}.xlxs': '{sheet_name}'. "
                f"Allowed names are 'Groups', 'Individuals', 'Tab*', Fig*'"
            )

    # 'Groups' sheet must exist
    if "Groups" not in sheet_names:
        logger.error(f"Groups sheet does not exist in sheets: '{sheet_names}'.")

    # Fix datatypes
    # for sheet_name in sheet_names:
    #     df = dfs[sheet_name]
    #     if sheet_name == "Groups":
    #         continue
    #     elif sheet_name == "Individuals":
    #         continue
    #     else:
    #         dfs[sheet_name] = df.astype({"group": str, "individual": str})

    # Validate individual sheets
    for sheet_name in sheet_names:
        df = dfs[sheet_name]
        # FIXME: transfer information from groups to individuals!

        if sheet_name == "Groups":
            validate_sheet_groups(df, study_id, sheet_name=sheet_name)
        elif sheet_name == "Individuals":
            validate_sheet_individuals(df, study_id, sheet_name=sheet_name)
        else:
            # add source information to the abundance table
            df.insert(loc=1, column="source", value=sheet_name)

            if len(df) == 0:
                # don't check empty data frames
                continue
            validate_sheet_data(df, study_id, sheet_name=sheet_name)

    # Validate information between sheets
    existing_groups = []
    if "Groups" in dfs:
        existing_groups = dfs["Groups"].name.unique()

    existing_individuals = []
    if "Individuals" in dfs:
        existing_individuals = dfs["Individuals"].name.unique()

    for sheet_name in sheet_names:
        if sheet_name in {"Groups", "Individuals"}:
            continue

        df = dfs[sheet_name]
        # group must exist in Groups
        for group in df.group.unique():
            if group is None:
                continue
            if group not in existing_groups:
                if not isinstance(group, str) and np.isnan(group):
                    pass
                else:
                    logger.error(
                        f"Group '{group}' from '{sheet_name}' does not exist in 'Groups'."
                    )

        # individual must exist in Individuals
        for individual in df.individual.unique():
            if individual is None:
                continue
            if individual not in existing_individuals:
                if not isinstance(individual, str) and np.isnan(individual):
                    pass
                else:
                    logger.error(
                        f"Individual '{individual}' from '{sheet_name}' does not exist in 'Individuals'."
                    )

    # console.print(dfs.keys())
    return dfs


def _check_required_columns(df: pd.DataFrame, required_columns: Iterable[str]) -> None:
    """Check required columns."""
    for field in required_columns:
        if field not in df.columns:
            logger.error(f"'{field}' column missing'.")


def _check_restricted_items(
    df: pd.DataFrame, restricted_items: Dict[str, List[str]]
) -> None:
    """Check for restricted items."""
    for key, allowed_values in restricted_items.items():
        if key in df.columns:
            for value in df[key].unique():
                if value not in allowed_values:
                    logger.error(f"{key} '{value}' must be in {allowed_values}")


def _check_measurement_types(df: pd.DataFrame, info_dict: Dict[str, Dict]) -> None:
    """Check values which are allowed for a given measurement type."""
    for _, row in df.iterrows():
        mtype = row.measurement_type
        if mtype not in info_dict:
            logger.error(f"Measurement type '{mtype}' is not supported.")
        else:
            info = info_dict[mtype]

            if info["dtype"] == "categorical":
                if not row.choice:
                    logger.error(f"'choice' must be set for '{mtype}' in '{row}'")
                else:
                    choices = info["choices"]
                    if choices and row.choice not in choices:
                        logger.error(
                            f"choice '{row.choice}' for '{mtype}' not in allowed choices: {choices} in '{row}'"
                        )
                for col in ["mean", "sd", "min", "max", "max", "median"]:
                    if col in row and row[col] and not np.isnan(row[col]):
                        logger.error(
                            f"'{col}' can not be set for categorical '{mtype}' in '{row}'."
                        )

            elif info["dtype"] == "numerical":
                for col in ["choice"]:
                    if (
                        row[col]
                        and not isinstance(row[col], str)
                        and not np.isnan(row[col])
                    ):
                        logger.error(
                            f"'{col}' can not be set for numerical '{mtype}' in '{row}'."
                        )
                for col in ["unit"]:
                    if not row[col]:
                        logger.error(
                            f"'{col}' is required for numerical '{mtype}' in '{row}'."
                        )
                    if not isinstance(row[col], str) and np.isnan(row[col]):
                        logger.error(
                            f"'{col}' is required for numerical '{mtype}' in '{row}'."
                        )


def _check_required_fields(df: pd.DataFrame, required_fields: Iterable[str]) -> None:
    """Check for required fields."""
    for field in required_fields:
        if field in df.columns:
            na_count = df[field].isnull().sum()
            if na_count > 0:
                logger.error(f"'{field}' is required, but some values are empty.")


def _check_numeric_fields(df: pd.DataFrame, numeric_fields: Iterable[str]) -> None:
    """Check that fields are numeric."""
    for field in numeric_fields:
        if field in df.columns:
            if not is_numeric_dtype(df[field]):
                logger.error(f"'{field}' must be numeric, but is {df[field].dtype}")


def validate_sheet(df: pd.DataFrame, study_id: str, sheet_name: str) -> None:
    """Validate generic sheet."""

    # study column must exist and must be study_id
    if "study" not in df.columns:
        logger.error(f"'study' column missing in '{study_id}_{sheet_name}'.")
    else:
        study_values = df.study.unique()
        if len(study_values) == 1:
            if study_values[0] != study_id:
                logger.error(
                    f"'study' column must be '{study_id}', but '{study_values[0]}'."
                )
        elif len(study_values) > 1:
            logger.error(
                f"'study' column must be '{study_id}', but multiple values found: "
                f"'{study_values}'"
            )


person_measurement_types = {
    "species": {
        "required": True,
        "dtype": "categorical",
        "choices": {
            "homo sapiens",
            "primates",
            "canis familiaris",
            "rattus norvegicus",
            "mammalia",
        },
    },
    "age": {"required": True, "dtype": "numerical", "unit": "yr", "min": 0, "max": 150},
    "weight": {
        "required": True,
        "dtype": "numerical",
        "unit": "kg",
        "min": 0,
        "max": 200,
    },
    "bmi": {
        "required": True,
        "dtype": "numerical",
        "unit": "kg/m^2",
        "min": 5,
        "max": 45,
    },
    "ethnicity": {
        "required": True,
        "dtype": "categorical",
        "choices": {
            "NR",
            "african",
            "african american",
            "arab american",
            "american indian",
            "asian",
            "korean",
            "japanese",
            "MF" "thai",
            "caucasian",
            "chinese",
            "hispanic",
            "white new zealanders",
            "asian indian",
            "afro trinidadians",
            "indo trinidadians",
            "black",
        },
    },
    "healthy": {
        "required": True,
        "dtype": "categorical",
        "choices": {"NR", "Y", "N", "YN"},
    },
    "sex": {
        "required": True,
        "dtype": "categorical",
        "choices": {"NR", "M", "F", "MF"},
    },
    "smoking": {
        "required": True,
        "dtype": "categorical",
        "choices": {"NR", "Y", "N", "YN"},
    },
    "alcohol": {
        "required": True,
        "dtype": "categorical",
        "choices": {"NR", "Y", "N", "YN"},
    },
    # not required
    "height": {
        "required": False,
        "dtype": "numerical",
        "unit": "cm",
        "min": 0,
        "max": 250,
    },
    "disease": {
        "required": False,
        "dtype": "categorical",
        "choices": {
            "NR",
            "hepatocellular cancer",
            "hepatocellular carcinoma",
            "hepatic cholangiocarcinoma",
            "adenocarcinoma",
            "cholangiocarcinoma",
            "cavernous hemangioma of liver",
            "intrahepatic stone",
            "gallbladder cancer",
            "papillary carcinoma",
            "diabetes",
            "cytostatic treatment",
            "gynecological cancer",
            "colorectal cancer",
            "metastatic carcinoma",
            "cholelithiasis",
        },
    },
    "cause of death": {"required": False, "dtype": "categorical", "choices": None},
    "medical history": {"required": False, "dtype": "categorical", "choices": None},
    "medication": {"required": False, "dtype": "categorical", "choices": None},
    "treatment": {"required": False, "dtype": "categorical", "choices": None},
    "anaesthetic": {"required": False, "dtype": "categorical", "choices": None},
    "liver pathology": {"required": False, "dtype": "categorical", "choices": None},
    "postmortem time": {"required": False, "dtype": "numerical", "unit": "min"},
}


def validate_sheet_groups(df: pd.DataFrame, study_id: str, sheet_name: str) -> None:
    """Validate group sheet."""
    console.print(f"validate_sheet_groups: {sheet_name}")
    validate_sheet(df=df, study_id=study_id, sheet_name=sheet_name)

    # check required columns
    required_columns = {
        "study",
        "name",
        "group_count",
        "parent",
        "measurement_type",
        "count",
        "substance",
        "choice",
        "mean",
        "sd",
        "min",
        "max",
        "unit",
        "comments",
    }
    _check_required_columns(df, required_columns=required_columns)

    restricted_items = {
        "measurement_type": list(person_measurement_types.keys()),
    }
    _check_restricted_items(df, restricted_items=restricted_items)

    _check_measurement_types(df, info_dict=person_measurement_types)

    # check for "all" group
    groups = df["name"].unique()
    if "all" not in groups:
        logger.error("Group 'all' missing")

    # check that parent is set
    for _, row in df.iterrows():
        if row["name"] == "all":
            continue
        if not row.parent or (isinstance(row.parent, float) and np.isnan(row.parent)):
            logger.error(f"'parent' group must be set for group '{row['name']}'")

    # check that count is set
    for _, row in df.iterrows():
        if not row["count"]:
            logger.error(f"'count' must be set in {row}")

    # FIXME: check the required fields for groups
    # console.print(df)
    # for group in df.name.unique():
    #     df_group = df[df.name == group]
    #     mtypes = df_group.measurement_type.unique()
    #     for mtype, mtype_info in person_measurement_types.items():
    #         required = mtype_info["required"]
    #         if required and not mtype in mtypes:
    #             logger.error(f"measurement_type '{mtype}' missing for group '{group}'")


def validate_sheet_individuals(
    df: pd.DataFrame, study_id: str, sheet_name: str
) -> None:
    """Validate individual sheet."""
    console.print(f"validate_sheet_individuals: {sheet_name}")
    validate_sheet(df=df, study_id=study_id, sheet_name=sheet_name)
    required_columns = [
        "study",
        "name",
        "group",
        "measurement_type",
        "substance",
        "choice",
        "value",
        "unit",
        "comments",
    ]
    _check_required_columns(df, required_columns=required_columns)

    restricted_items = {
        "measurement_type": list(person_measurement_types.keys()),
    }
    _check_restricted_items(df, restricted_items=restricted_items)

    _check_measurement_types(df, info_dict=person_measurement_types)

    # check that group is set
    for _, row in df.iterrows():
        if not row["group"]:
            logger.error(f"'group' must be set in {row}")


def validate_sheet_data(df: pd.DataFrame, study_id: str, sheet_name: str) -> None:
    """Validate data sheet."""
    console.print(f"validate_sheet_data: {sheet_name}")
    validate_sheet(df=df, study_id=study_id, sheet_name=sheet_name)

    # check required columns
    required_columns = [
        "study",
        "count",
        "measurement_type",
        "protein",
        "tissue",
        "method",
        "value",
        "mean",
        "sd",
        "median",
        "min",
        "max",
        "cv",
        "cv_unit",
        "mad",
        "unit",
        "comments",
    ]
    _check_required_columns(df, required_columns=required_columns)

    restricted_items = {
        "tissue": [
            "liver microsomes",
            "liver plasma membrane",
            "liver membrane",
            "kidney membrane",
            "hepatocyte",
            "jejunum",
            "duodenum",
        ],
        "measurement_type": ["abundance"],
    }
    _check_restricted_items(df, restricted_items=restricted_items)

    required_fields = [
        "study",
        "count",
        "measurement_type",
        "protein",
        "tissue",
        "unit",
    ]
    _check_required_fields(df, required_fields=required_fields)

    # test that either group or individual is set
    if "group" in df.columns and "individual" in df.columns:
        # check that both are zero
        item_count = (~df["group"].isnull()).sum() + (~df["individual"].isnull()).sum()
        if item_count != len(df):
            logger.error(
                "either 'individual' or 'group' must be set, but some values are empty or have both."
            )

    # check that fields are numeric
    numeric_fields = [
        "value",
        "mean",
        "sd",
        "median",
        "min",
        "max",
        "cv",
        "mad",
        "count",
    ]
    _check_numeric_fields(df, numeric_fields=numeric_fields)


def merge_tables(
    data: Dict[str, pd.DataFrame], sids: List[str], excel_path: Optional[Path]
) -> pd.DataFrame:
    """Merge tables in a single table."""

    dfs: List[pd.DataFrame] = []
    for sid in sids:
        df = data[sid]
        df["sid"] = sid
        dfs.append(data[sid])

    df_merged = pd.concat(dfs)
    df_merged.to_excel(excel_path, sheet_name="abundance", index=False)
    return df_merged


def process_all_data(
    data_xlsx: Path,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Process all study data."""
    # get the study ids from the study_overview table
    df_studies = load_study_overview()
    study_ids = sorted(df_studies.study.values)

    excluded_ids: List[str] = []
    study_ids = [sid for sid in study_ids if sid not in excluded_ids]
    # study_ids = ["Achour2014"]

    console.print(study_ids)

    # try to load individual study
    # study_id = study_ids[0]
    # data = dataio.load_data_from_study(study_id)

    # try to load all studies
    data: Dict[str, Dict[str, pd.DataFrame]] = load_data(study_ids=study_ids)

    # at the end of the data reading we have 3 large tables
    # merge table:
    # 1. Groups data
    # 2. Individual data
    # 3. Abundance data
    dfs_groups = []
    dfs_individuals = []
    dfs_abundance = []
    for _, data_dict in data.items():
        for sheet_name, df in data_dict.items():
            if sheet_name == "Groups":
                dfs_groups.append(df)
            elif sheet_name == "Individuals":
                dfs_individuals.append(df)
            else:
                dfs_abundance.append(df)

    df_groups = pd.concat(dfs_groups)
    df_individuals = pd.concat(dfs_individuals)
    df_abundance = pd.concat(dfs_abundance)

    # unit normalization
    for unit in df_abundance["unit"].unique():
        if unit != "pmol/mg":
            if unit == "fmol/mug":
                factor_unit = 1.0
            elif unit == "fmol/µg":
                factor_unit = 1.0
            elif unit == "amol/mug":
                factor_unit = 1000.0  # FIXME: check this
            elif unit == "nmol/mg":
                factor_unit = 1000.0
            elif unit == "fmol/g":
                factor_unit = 1000.0  # FIXME: check this
            else:
                logger.error(f"Unsupported unit: '{unit}'")
                continue

            row_index = df_abundance.unit == unit
            for key in [
                "value",
                "mean",
                "sd",
                "se",
                "median",
                "min",
                "max",
                "mean_pm_sd",
            ]:
                df_abundance.loc[row_index, key] = (
                    df_abundance[row_index][key] * factor_unit
                )
            df_abundance.loc[row_index, "unit"] = "pmol/mg"

    # calculate SD from SE
    row_index = pd.isnull(df_abundance["sd"]) & ~pd.isnull(df_abundance["se"])
    df_abundance.loc[row_index, "sd"] = df_abundance[row_index]["se"] * np.sqrt(
        df_abundance[row_index]["count"]
    )

    # calculate SD from CV
    row_index = (
        pd.isnull(df_abundance["sd"])
        & ~pd.isnull(df_abundance["cv"])
        & (df_abundance["cv_unit"] == "percent")
    )
    df_abundance.loc[row_index, "sd"] = (
        df_abundance[row_index]["cv"] / 100.0 * df_abundance[row_index]["mean"]
    )

    # set value for individuals from mean
    row_index = (
        (pd.isnull(df_abundance["value"]))
        & (~pd.isnull(df_abundance["individual"]))
        & (~pd.isnull(df_abundance["mean"]))
    )
    df_abundance.loc[row_index, "value"] = df_abundance[row_index]["mean"]
    # remove mean
    df_abundance.loc[row_index, "mean"] = np.NaN
    df_abundance.loc[row_index, "sd"] = np.NaN

    # filter empty rows
    df_abundance.reset_index(inplace=True, drop=True)
    row_index = (pd.isnull(df_abundance["value"])) & (pd.isnull(df_abundance["mean"]))
    df_abundance.drop(df_abundance[row_index].index, inplace=True)

    # remove unnecessary columns
    for col in [
        "heterogeneity",
        "q",
        "r2",
        "r2_unit",
        "mean_pm_sd",
        "percent_total_cyp",
        "max_over_min",
        "max/min",
        "variance",
        "q1",
        "q3",
        "logmean",
        "logmean_unit",
        "sample_name",
        "vmax",
        "vmax_unit",
        # additional columns
        "mad",
        "cv",
        "cv_unit",
        "median",
        "se",
        "method",
        "min",
        "max",
    ]:
        if col in df_abundance.columns:
            df_abundance.drop(columns=[col], inplace=True)

    with pd.ExcelWriter(data_xlsx) as writer:
        df_groups.to_excel(writer, sheet_name="Groups", index=False)
        df_individuals.to_excel(writer, sheet_name="Individuals", index=False)
        df_abundance.to_excel(writer, sheet_name="Abundance", index=False)

    return df_groups, df_individuals, df_abundance


def filter_data(
    data_raw_xlsx: Path,
    data_xlsx: Path,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Filter data.

    Reduce to data with sufficient data points.
    """

    df_groups = pd.read_excel(data_raw_xlsx, sheet_name="Groups")
    df_individuals = pd.read_excel(data_raw_xlsx, sheet_name="Individuals")
    df_abundance = pd.read_excel(data_raw_xlsx, sheet_name="Abundance")

    # filter proteins which are not in human
    count_start = len(df_abundance)

    proteins_nonhuman = df_abundance[
        df_abundance.protein.str.contains("_")
    ].protein.unique()
    df_abundance = df_abundance[~df_abundance.protein.str.contains("_")]

    console.print(
        f"Filtered '{count_start - len(df_abundance)}' non-human entries: '{proteins_nonhuman}'"
    )

    # Filter minimal count of proteins
    value_cutoff = 10
    proteins = df_abundance.protein.unique()
    for pid in proteins:
        if df_abundance[df_abundance.protein == pid].value.count() < value_cutoff:
            console.print(f"Filtered protein '{pid}' with < {value_cutoff} values.")
            df_abundance = df_abundance[~(df_abundance.protein == pid)]

    with pd.ExcelWriter(data_xlsx) as writer:
        df_groups.to_excel(writer, sheet_name="Groups", index=False)
        df_individuals.to_excel(writer, sheet_name="Individuals", index=False)
        df_abundance.to_excel(writer, sheet_name="Abundance", index=False)

    return df_groups, df_individuals, df_abundance


def process_merge_sheets(data_xlsx: Path, merged_xlsx: Path) -> None:
    """Merge group/individual information with abundance.

    - [ ] format sheets
    - [ ] create merged table with subject/group information
    - [ ] calculate group characteristica from individual characteristica and viceversa
    """
    df_groups = pd.read_excel(data_xlsx, sheet_name="Groups")
    df_groups.insert(
        0, column="sid", value=(df_groups["study"] + "_" + df_groups["name"])
    )

    df_individuals = pd.read_excel(data_xlsx, sheet_name="Individuals")
    df_individuals.insert(
        0, column="sid", value=(df_individuals["study"] + "_" + df_individuals["name"])
    )

    df_abundance = pd.read_excel(data_xlsx, sheet_name="Abundance")

    # get individual abundance
    data_individuals: pd.DataFrame = df_abundance.loc[
        ~pd.isnull(df_abundance["value"]), :
    ]
    data_individuals = data_individuals.copy()
    data_individuals.drop(columns=["group", "mean", "sd", "comments"], inplace=True)
    data_individuals.insert(
        0,
        column="sid",
        value=(data_individuals["study"] + "_" + data_individuals["individual"]),
    )

    # get group abundance
    data_groups: pd.DataFrame = df_abundance.loc[~pd.isnull(df_abundance["mean"]), :]
    data_groups = data_groups.copy()
    data_groups.drop(columns=["individual", "value", "comments"], inplace=True)
    data_groups.insert(
        0, column="sid", value=(data_groups["study"] + "_" + data_groups["group"])
    )

    # TODO: processing of group/individual information still missing
    individual_data = {}
    mtypes = {"species", "sex", "age", "bmi", "smoking", "alcohol", "ethnicity"}
    for sid in df_individuals.sid.unique():
        individual_data[sid] = {
            "sid": sid,
            "species": "homo sapiens",
            "sex": "NR",
            "age": np.NaN,
            "age_group": "NR",
            "bmi": np.NaN,
            "bmi_group": "NR",
            "smoking": "NR",
            "alcohol": "NR",
            "ethnicity": "NR",
        }

    # sex, age, bmi, smoking, alcohol, ethnicity
    for _, row in df_individuals.iterrows():
        mtype = row.measurement_type
        if mtype not in mtypes:
            continue

        if mtype in {"species", "sex", "smoking", "alcohol", "ethnicity"}:
            individual_data[row.sid][mtype] = row.choice
        elif mtype in {"age", "bmi"}:
            individual_data[row.sid][mtype] = row.value

            # Major adult BMI classifications are
            # - underweight: bmi < 18.5 kg/m2
            # - normal weight: 18.5 <= bmi < 25)
            # - overweight: 25 <= bmi < 30)
            # - obese: 30 <= bmi
            if mtype == "bmi":
                if row.value < 18.5:
                    bmi_group = "underweight"
                elif 18.5 <= row.value < 25:
                    bmi_group = "normal weight"
                elif 25 <= row.value < 30:
                    bmi_group = "overweight"
                elif 30 <= row.value:
                    bmi_group = "obese"
                elif np.isnan(row.value):
                    bmi_group = "NR"

                individual_data[row.sid]["bmi_group"] = bmi_group

            # Major age classifications are
            if mtype == "age":
                if row.value < 18:
                    age_group = "adolescent"
                elif 18 <= row.value < 35:
                    age_group = "adult"
                elif 35 <= row.value < 65:
                    age_group = "middle aged"
                elif 65 <= row.value:
                    age_group = "elderly"
                elif np.isnan(row.value):
                    age_group = "NR"

                individual_data[row.sid]["age_group"] = age_group

    df_characteristica = pd.DataFrame(individual_data.values())
    data_individuals = data_individuals.merge(df_characteristica, on="sid")

    # process correlation information for group data
    corr_data: Dict[str, Dict[str, Any]] = {}
    corr_count: Dict[str, int] = {}
    for _, row in data_groups.iterrows():
        # key = f"{row['study']}_{row['source']}_{row['group']}_{row['tissue']}"
        key = f"{row['study']}_{row['group']}_{row['tissue']}"
        if key in corr_data:
            d = corr_data[key]
        else:
            d = {
                "study": row["study"],
                # "source": row["source"],
                "group": row["group"],
                "tissue": row["tissue"],
            }
        d[row["protein"]] = row["mean"]
        d[f'{row["protein"]}_sd'] = row["sd"]
        corr_count[key] = corr_count.get(key, 0) + 1
        corr_data[key] = d

    # filter data to entries with at least two proteins
    corr_data = {k: v for k, v in corr_data.items() if corr_count[k] > 1}
    correlation_groups = pd.DataFrame(corr_data.values())
    protein_columns = [
        c
        for c in correlation_groups.columns
        if c not in ["study", "source", "group", "tissue"]
    ]
    # filter to proteins with uniprot
    columns = ["study", "group", "tissue"] + sorted(protein_columns)
    correlation_groups = correlation_groups[columns]

    # process correlation information for individual data
    corr_data = {}
    corr_count = {}
    for _, row in data_individuals.iterrows():
        key = f"{row['study']}_{row['individual']}_{row['tissue']}"
        if key in corr_data:
            d = corr_data[key]
        else:
            d = {
                "study": row["study"],
                "individual": row["individual"],
                "tissue": row["tissue"],
            }
        d[row["protein"]] = row["value"]
        corr_count[key] = corr_count.get(key, 0) + 1
        corr_data[key] = d

    corr_data = {k: v for k, v in corr_data.items() if corr_count[k] > 1}
    correlation_individuals = pd.DataFrame(corr_data.values(), index=corr_data.keys())
    # order columns
    protein_columns = [
        c
        for c in correlation_individuals.columns
        if c not in ["study", "source", "individual", "tissue"]
    ]

    proteins = protein_info.get_proteins(df_abundance, uniprot=True)
    protein_columns = [p for p in protein_columns if p in set(proteins)]
    columns = ["study", "individual", "tissue"] + sorted(protein_columns)
    correlation_individuals = correlation_individuals[columns]
    # at least 5 values required
    for protein_id in protein_columns:
        if correlation_individuals[protein_id].count() < 10:
            console.log(f"Not enough values, drop: {protein_id}")
            correlation_individuals.drop(columns=[protein_id], inplace=True)

    protein_categories = protein_info.get_protein_categories(proteins)

    # write data
    with pd.ExcelWriter(merged_xlsx) as writer:
        data_groups.to_excel(writer, sheet_name="group_data", index=False)
        data_individuals.to_excel(writer, sheet_name="individual_data", index=False)
        correlation_groups.to_excel(writer, sheet_name="group_correlation", index=False)
        correlation_individuals.to_excel(
            writer, sheet_name="individual_correlation", index=False
        )

        meta_cols = ["study", "individual", "tissue"]
        for category, protein_ids in protein_categories.items():
            protein_cols = [
                c for c in correlation_individuals.columns if c in protein_ids
            ]
            # find empty rows
            df_tmp_proteins = correlation_individuals[protein_cols]
            empty_row_idx = df_tmp_proteins.isna().all(axis="columns")

            all_cols = meta_cols + protein_cols
            df_tmp = correlation_individuals[all_cols]
            df_tmp = df_tmp[~empty_row_idx]
            df_tmp.to_excel(writer, sheet_name=f"individual_correlation_{category}")

    return None


def prepare_normalization_sheets(normalization_xlsx: Path, merged_xlsx: Path) -> None:
    """Prepare data for normalization."""
    console.rule(title="XLSX for normalization", align="left", style="white")
    df_group_data: pd.DataFrame = pd.read_excel(merged_xlsx, sheet_name="group_data")
    df_individual_data: pd.DataFrame = pd.read_excel(
        merged_xlsx, sheet_name="individual_data"
    )

    # FIXME: bugfix of additional column, this should never be written
    del df_group_data["protein.1"]
    del df_individual_data["protein.1"]

    # FIXME: calculate a dset_id for now (improve this)
    for df in [df_group_data, df_individual_data]:
        df.insert(1, column="dset_id", value=(df["study"] + "_" + df["source"]))

    # FIXME: format the excel automatically
    with pd.ExcelWriter(normalization_xlsx) as writer:
        df_individual_data.to_excel(writer, sheet_name="individual_data", index=False)
        df_group_data.to_excel(writer, sheet_name="group_data", index=False)

    console.print(normalization_xlsx)


def run_all_dataio() -> None:
    """Run all processing steps."""
    from protein_distribution import (
        DATA_MERGED_XLSX,
        DATA_NORMALIZATION_XLSX,
        DATA_RAW_XLSX,
        DATA_XLSX,
    )

    # process all data
    process_all_data(data_xlsx=DATA_RAW_XLSX)

    # apply filters
    df_groups, df_individuals, df_abundance = filter_data(
        data_raw_xlsx=DATA_RAW_XLSX, data_xlsx=DATA_XLSX
    )

    # show categories for data
    proteins = protein_info.get_proteins(df_abundance, uniprot=True)
    protein_categories = protein_info.get_protein_categories(proteins)
    console.rule(style="white")
    console.print(protein_categories)
    console.rule(style="white")

    # merge sheets for data
    process_merge_sheets(data_xlsx=DATA_XLSX, merged_xlsx=DATA_MERGED_XLSX)
    # subset of data for normalization
    prepare_normalization_sheets(
        normalization_xlsx=DATA_NORMALIZATION_XLSX, merged_xlsx=DATA_MERGED_XLSX
    )


if __name__ == "__main__":
    run_all_dataio()



"""
Testing of differences & multiple testing correction.

1. Independence of observations (a.k.a. no autocorrelation):
The observations/variables you include in your test are not related
(for example, multiple measurements of a single test subject are not independent,
while measurements of multiple different test subjects are independent).

=> measurement of multple different test sujects are indepenent.

2. Homogeneity of variance: the variance within each group being compared
is similar among all groups. If one group has much more variation than others,
it will limit the test’s effectiveness.

=> Before conducting the two-sample T-Test we need to find if the given data groups
have the same variance. If the ratio of the larger data groups to the small data
group is less than 4:1 then we can consider that the given data groups have equal variance.


3. Normality of data: the data follows a normal distribution (a.k.a. a bell curve).
This assumption applies only to quantitative data.

=> if 2 and 3 are not valid then we have to use nonparametric tests

Independent t-test:

p-values: * = 0.05, ** = 0.01, *** = 0.001

Multiple-testing correction:
https://www.statsmodels.org/stable/generated/statsmodels.stats.multitest.multipletests.html

2: groups testing (t-test)
- "sex", "smoking", "alcohol", "age_group"
3: groups testing (anova)
- "bmi_group"
no testing:
- ethnicities.
"""
from typing import Any, Dict

import numpy as np
import pandas as pd
import scipy.stats as stats

from protein_distribution.console import console
from protein_distribution.log import get_logger


logger = get_logger(__name__)


stratification_info = {
    "sex": {"groups": ["M", "F"], "colors": ["tab:blue", "tab:red"]},
    "smoking": {"groups": ["Y", "N"], "colors": ["tab:brown", "tab:gray"]},
    "alcohol": {"groups": ["Y", "N"], "colors": ["tab:olive", "tab:gray"]},
    "ethnicity": {
        "groups": ["caucasian", "african american", "hispanic"],
        "colors": ["tab:blue", "tab:orange", "tab:green"],
    },
    "age_group": {
        "groups": ["middle aged", "elderly"],
        "colors": ["#fdbb84", "#e34a33"],
    },
    "bmi_group": {
        "groups": ["normal weight", "overweight", "obese"],
        "colors": ["#d7b5d8", "#df65b0", "#ce1256"],
    },
}


def process_stratification_data(
    data: pd.DataFrame, stratification_key: str
) -> pd.DataFrame:
    """Process stratification data.

    existing groups and filtered groups:

    sex: [M, F, NR] -> [M, F]
    smoking [Y, N, NR] -> [Y, N]
    alcohol [Y, N, NR] -> [Y, N]
    ethnicity [caucasian, hispanic, african american, NR] -> [caucasian, african american, hispanic]
    age_group [adolescent, young, middle aged, elderly, NR] -> [middle aged, elderly]
    bmi_group [underweight, normal weight, overweight, obese, NR] -> [normal weight, overweight, obese]
    """

    if stratification_key not in stratification_info:
        raise ValueError(f"Unsupported stratification: {stratification_key}")

    # subset of abundance data for individual protein
    df_protein = data[~pd.isnull(data.value)]

    # filter out NR values
    df_protein = df_protein[~(df_protein[stratification_key] == "NR")]

    # filter out bmi_group: "underweight"
    if stratification_key == "bmi_group":
        df_protein = df_protein[~(df_protein[stratification_key] == "underweight")]
    # filter out age_group: "adolescent
    if stratification_key == "age_group":
        df_protein = df_protein[~(df_protein[stratification_key] == "adolescent")]

    # A minimum of 5 data points for every group required
    proteins_with_data = []
    groups = stratification_info[stratification_key]["groups"]
    for p in df_protein.protein.unique():
        if stratification_key == "ethnicity":
            include = False
        else:
            include = True

        for group in groups:
            df_group = df_protein[
                (df_protein.protein == p) & (df_protein[stratification_key] == group)
            ]

            if stratification_key == "ethnicity":
                # at least one group >= 10 subjects
                if len(df_group) >= 10:
                    include = True
                    break
            else:
                # all groups >= 10 subjects
                if len(df_group) < 10:
                    # console.print(f"Filtered protein '{p}' due to '{stratification_key}' = {group}")
                    include = False
                    break

        if include:
            proteins_with_data.append(p)

    # console.print("proteins with data", len(proteins_with_data) / len(df_protein.protein.unique()))

    # filter by proteins with sufficient data
    df_protein = df_protein[df_protein["protein"].isin(proteins_with_data)]

    return df_protein


def equality_of_variance(dfs: Dict[str, pd.DataFrame]) -> Dict:
    """Test equality of variance of groups.

    can be tested using Levene’s, Bartlett’s, or Brown-Forsythe test
    """

    variances = {group: np.var(df.value) for group, df in dfs.items()}
    variances_values = variances.values()
    var_min = min(variances_values)
    var_max = max(variances_values)
    var_ratio = var_max / var_min

    # Test that ratio < 4:1
    var_equal: bool = var_ratio <= 4.0
    # if not var_equal:
    #    logger.error(f"Variance ratio > 4:1, {var_ratio:.2f}")

    return {
        "variance ratio": var_ratio,
        "variance equal": var_equal,
    }


def significance_for_pvalue(pvalue: float) -> str:
    """Significance marker for pvalue."""
    if pvalue <= 0.001:
        significance = "***"
    elif pvalue <= 0.01:
        significance = "**"
    elif pvalue <= 0.05:
        significance = "*"
    else:
        significance = " "
    return significance


def two_sample_ttest(dfs: Dict[str, pd.DataFrame]) -> Dict[str, Any]:
    """Calculate two-sample t-test."""
    # perform testing of variance
    var_results = equality_of_variance(dfs_ttest)

    # group data
    data = [df.value.values for df in dfs.values()]

    # Perform the two sample t-test with equal variances
    t_results = stats.ttest_ind(a=data[0], b=data[1], equal_var=True)

    pvalue = t_results.pvalue
    significance = significance_for_pvalue(pvalue)
    # console.print(f"p-value={t_results.pvalue}, df={t_results.df}, u significance: '{significance}'")
    return {
        "method": "t-test",
        "pvalue": pvalue,
        "df": t_results.df,
        "sig": significance,
        **var_results,
    }


def two_sample_mannwitneyu(dfs: Dict[str, pd.DataFrame]) -> Dict[str, Any]:
    """Calculate two-sample Mann-Whitney test (unpaired).

    The most popular alternative is the Mann-Whitney test, which is done with
    the stats::wilcox.test function in R
    (http://stat.ethz.ch/R-manual/R-devel/library/stats/html/wilcox.test.html).

    The Wilcoxon signed-rank test tests the null hypothesis that two related paired
    samples come from the same distribution. In particular, it tests whether the
    distribution of the differences x - y is symmetric about zero. It is a non-parametric
    version of the paired T-test.
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.wilcoxon.html
    """
    # group data
    data = [df.value.values for df in dfs.values()]

    # Perform the two sample t-test with equal variances
    test_results = stats.mannwhitneyu(x=data[0], y=data[1])

    pvalue = test_results.pvalue
    significance = significance_for_pvalue(pvalue)
    # console.print(f"p-value={t_results.pvalue}, df={t_results.df}, u significance: '{significance}'")
    return {
        "method": "Mann-Whitney",
        "pvalue": pvalue,
        "sig": significance,
    }


def kruskal_wallis(dfs: Dict[str, pd.DataFrame]) -> Dict[str, Any]:
    """Kruskal Wallis significance test of data.

    A Kruskal-Wallis test is used to determine whether or not there is a
    statistically significant difference between the medians of three or
    more independent groups.

    This test is the nonparametric equivalent of the one-way ANOVA and is typically
    used when the normality assumption is violated.

    The Kruskal-Wallis test does not assume normality in the data and is much less
    sensitive to outliers than the one-way ANOVA.

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.kruskal.html

    :return:
    """
    # group data
    data = [df.value.values for df in dfs.values()]

    # Perform kruskall wallis
    test_results = stats.kruskal(*data)

    pvalue = test_results.pvalue
    significance = significance_for_pvalue(pvalue)

    return {
        "method": "Kruskal-Wallis",
        "pvalue": pvalue,
        "sig": significance,
    }


if __name__ == "__main__":
    from protein_distribution import DATA_MERGED_XLSX, DATA_XLSX, RESULTS_DIR
    from protein_distribution.protein_info import get_protein_categories, get_proteins

    df_abundance = pd.read_excel(DATA_XLSX, sheet_name="Abundance")
    del df_abundance["comments"]
    df_abundance["sid"] = df_abundance["study"] + "_" + df_abundance["source"]

    proteins = get_proteins(df_abundance, uniprot=True)
    protein_categories = get_protein_categories(proteins)

    df_individual_data = pd.read_excel(DATA_MERGED_XLSX, sheet_name="individual_data")

    ttest_results = []
    for category in ["cyp", "ugt", "slc", "abc"]:
        df_category = df_individual_data[
            df_individual_data.protein.isin(protein_categories[category])
        ]

        for stratification_key in [
            "sex",
            "smoking",
            "alcohol",
            "ethnicity",
            "age_group",
            "bmi_group",
        ]:
            groups = stratification_info[stratification_key]["groups"]
            df_proteins = process_stratification_data(
                df_category, stratification_key=stratification_key
            )
            # print(df_proteins)

            proteins = df_proteins.protein.unique()
            for protein_id in proteins:
                dfs_ttest = {}
                for group in groups:
                    # console.print(f"protein: {protein_id}, key: {stratification_key}, group: {group}")
                    df = df_proteins[
                        (df_proteins.protein == protein_id)
                        & (df_proteins[stratification_key] == group)
                    ]
                    dfs_ttest[group] = df

                # t-tests/wilcoxon/Mann-Whitney
                if stratification_key in ["sex", "alcohol", "age_group", "smoking"]:
                    # test_results = two_sample_ttest(dfs_ttest)
                    test_results = two_sample_mannwitneyu(dfs_ttest)

                # anova (Kruskall Wallis)
                elif stratification_key in ["bmi_group"]:
                    test_results = kruskal_wallis(dfs_ttest)

                elif stratification_key in ["ethnicity"]:
                    # not tested
                    continue

                ttest_results.append(
                    {
                        "category": category,
                        "protein": protein_id,
                        "variable": stratification_key,
                        **test_results,
                    }
                )

    df_ttest = pd.DataFrame(ttest_results)
    df_ttest.sort_values(by="pvalue", inplace=True)

    # multiple testing correction
    from statsmodels.stats.multitest import multipletests

    pvals = df_ttest.pvalue.values
    (
        reject,
        pvals_corrected,
        _,
        _,
    ) = multipletests(
        pvals,
        alpha=0.05,  # FWER, family-wise error rate, e.g. 0.1
        method="fdr_bh",  # hs,
        maxiter=1,
        is_sorted=False,
        returnsorted=False,
    )
    df_ttest["pvalue corrected"] = pvals_corrected
    df_ttest["sig corrected"] = [
        significance_for_pvalue(pval) for pval in pvals_corrected
    ]

    console.rule(style="white")
    console.print(df_ttest.to_string())
    console.rule(style="white")

    table_significance = RESULTS_DIR / "table_stratification.xlsx"
    with pd.ExcelWriter(table_significance) as writer:
        for category in ["cyp", "ugt", "slc", "abc"]:
            df = df_ttest[df_ttest.category == category]
            df.to_excel(writer, sheet_name=category, index=False)



"""Create tables with correlation information.

Correlation coefficient, p-value, significance.
sig.level = c(0.001, 0.01, 0.05),
"""

from typing import Dict

# ------------------------------------------------------------------------------
import matplotlib
import matplotlib.axes
import matplotlib.pyplot
import pandas as pd
from matplotlib import pyplot as plt

from protein_distribution import RESULTS_DIR
from protein_distribution.console import console


SMALL_SIZE = 12
MEDIUM_SIZE = 15
BIGGER_SIZE = 25

matplotlib.rc("font", size=SMALL_SIZE)  # controls default text sizes
matplotlib.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
matplotlib.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
matplotlib.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
matplotlib.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
matplotlib.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title
# ------------------------------------------------------------------------------


def correlation_volcano_plot(
    category: str, cutoff_pvalue: float = 0.05, cutoff_correlation: float = 0.5
) -> None:
    """Volcano plot of the correlation coeffient."""

    # load correlations and p-values
    df_M = pd.read_table(
        RESULTS_DIR / "correlation" / f"M_individual_correlation_{category}.csv",
        sep=",",
    )
    df_M.set_index(keys="Unnamed: 0", inplace=True)

    df_p = pd.read_table(
        RESULTS_DIR / "correlation" / f"p_individual_correlation_{category}.csv",
        sep=",",
    )
    df_p.set_index(keys="Unnamed: 0", inplace=True)
    console.rule(align="left", title=category)

    columns = df_M.columns

    # plot all data points
    labels = []
    correlations = []
    pvalues = []
    for kx, protein_x in enumerate(columns):
        for ky, protein_y in enumerate(columns):
            if kx >= ky:
                continue
            if category in {"cyp", "ugt"}:
                label = f"{protein_x[3:]}-{protein_y[3:]}"
            else:
                label = f"{protein_x}~{protein_y}"
            correlation = df_M.iloc[kx, ky]
            pvalue = df_p.iloc[kx, ky]

            labels.append(label)
            correlations.append(correlation)
            pvalues.append(pvalue)

    df_data = pd.DataFrame(
        {"label": labels, "correlation": correlations, "pvalue": pvalues}
    )
    df_data.sort_values(by=["correlation"])
    console.print(df_data)

    # create plot
    ax: matplotlib.axes.Axes
    fig, ax = plt.subplots(nrows=1, ncols=1, dpi=300, figsize=(10, 8))
    ax.set_xlabel("Correlation", fontweight="bold")
    ax.set_ylabel("p-value", fontweight="bold")
    if category == "cyp":
        title = "Cytochrome P450 (CYP)"
    elif category == "ugt":
        title = "UDP-glucuronosyltransferase (UGT)"
    else:
        title = category
    # fig.suptitle(title)
    ax.set_title(title, fontweight="bold", fontsize=16)

    # lines
    line_kwargs = {"linewidth": 1.0, "linestyle": "-", "color": "black"}
    ax.axhline(y=cutoff_pvalue, **line_kwargs)
    ax.axvline(x=-cutoff_correlation, **line_kwargs)
    ax.axvline(x=+cutoff_correlation, **line_kwargs)

    # points with text annotations
    for k, label in enumerate(labels):
        pvalue = pvalues[k]
        correlation = correlations[k]

        kwargs = {
            "color": "darkgrey",
            "alpha": 0.5,
        }
        if correlation > cutoff_correlation and pvalue < cutoff_pvalue:
            kwargs = {
                "color": "tab:blue",
                "alpha": 0.9,
            }
        elif correlation < -cutoff_correlation and pvalue < cutoff_pvalue:
            kwargs = {
                "color": "tab:red",
                "alpha": 0.9,
            }

        ax.plot(
            correlation,
            pvalue,
            # linestyle="",
            marker="o",
            markersize=7,
            linestyle="",
            # markeredgecolor="black",
            **kwargs,
        )

        if abs(correlation) > cutoff_correlation and pvalue < cutoff_pvalue:
            ax.annotate(
                text=label,
                xy=(correlation, pvalue),
                fontsize="xx-small",
                fontweight="bold",
                # backgroundcolor="white",
                rotation=0,
            )
    ax.set_yscale("log")
    ax.grid(True)
    ylim = ax.get_ylim()
    ax.set_ylim(ylim[1], ylim[0])

    fig.savefig(
        RESULTS_DIR / "correlation" / f"volcano_{category}.png", bbox_inches="tight"
    )
    fig.savefig(RESULTS_DIR / "correlation" / f"volcano_{category}.svg")

    plt.show()


def correlation_tables() -> None:
    """Create correlation tables."""
    dfs_correlation: Dict[str, pd.DataFrame] = {}
    for category in categories:
        df_M = pd.read_table(
            RESULTS_DIR / "correlation" / f"M_individual_correlation_{category}.csv",
            sep=",",
        )
        df_p = pd.read_table(
            RESULTS_DIR / "correlation" / f"p_individual_correlation_{category}.csv",
            sep=",",
        )

        dfs_correlation[f"{category}_M"] = df_M
        dfs_correlation[f"{category}_p"] = df_p

    # write results
    table_correlations = RESULTS_DIR / "table_correlations.xlsx"

    # write correlation value & p-value in table for categories
    with pd.ExcelWriter(table_correlations) as writer:
        for key, df in dfs_correlation.items():
            df.to_excel(writer, sheet_name=key, index=False)


if __name__ == "__main__":
    console.rule(style="white")
    categories = ["cyp", "ugt", "slc", "abc"]

    # create tables
    correlation_tables()

    # volcano plots
    for category in categories:
        correlation_volcano_plot(category=category)



"""Test that all resources for the app are available."""

from protein_distribution import (
    DATA_MERGED_XLSX,
    DATA_RAW_XLSX,
    DATA_XLSX,
    UNIPROT_PATH,
)


def test_app1() -> None:
    """Test that all resources can be loaded for the app."""
    assert UNIPROT_PATH.exists()


def test_app2() -> None:
    """Test that all resources can be loaded for the app."""
    assert DATA_RAW_XLSX.exists()


def test_app3() -> None:
    """Test that all resources can be loaded for the app."""
    assert DATA_MERGED_XLSX.exists()


def test_app4() -> None:
    """Test that all resources can be loaded for the app."""
    assert DATA_XLSX.exists()



"""Create overview tables."""

from pathlib import Path
from typing import Any, Dict, Iterable, List

import numpy as np
import pandas as pd
from scipy import stats

from protein_distribution import (
    DATA_MERGED_XLSX,
    DATA_XLSX,
    RESULTS_DIR,
    log,
    protein_info,
)
from protein_distribution.console import console


logger = log.get_logger(__name__)


def create_study_table(
    proteins: List[str],
    df: pd.DataFrame,
    df_groups: pd.DataFrame,
    df_individuals: pd.DataFrame,
) -> Dict[str, Dict[str, str]]:
    """Create study overview table for given list of proteins."""
    data: Dict[str, Dict[str, str]] = {}
    all_studies = df.study.unique()
    for protein in proteins:
        df_protein = df[df.protein == protein]
        studies = df_protein.study.unique()
        info = {}
        for study in all_studies:
            # get group data
            dfg = df_groups.loc[
                (df_groups.protein == protein) & (df_groups.study == study), :
            ]
            n_groups = len(dfg)

            # get individual data
            dfi = df_individuals.loc[
                (df_individuals.protein == protein) & (df_individuals.study == study), :
            ]
            n_individuals = len(dfi)

            if study in studies:  # ✓
                str_groups = f"{n_groups if n_groups else ' '}".rjust(2)
                str_individuals = f"{n_individuals if n_individuals else ' '}".ljust(2)
                info[study] = f"{str_groups}|{str_individuals}"
            else:
                info[study] = ""
        data[protein] = info

    return data


def create_study_tables(
    tables_xlsx: Path,
    drop_empty: bool = True,
    sort_by_count: bool = True,
) -> None:
    """Create all study overview tables."""
    # Load data
    df_abundance = pd.read_excel(DATA_XLSX, sheet_name="Abundance")
    df_groups = pd.read_excel(DATA_MERGED_XLSX, sheet_name="group_data")
    df_individuals = pd.read_excel(DATA_MERGED_XLSX, sheet_name="individual_data")

    proteins = protein_info.get_proteins(df_abundance, uniprot=True)
    protein_categories = protein_info.get_protein_categories(proteins)

    # create study tables
    dfs_table: Dict[str, pd.DataFrame] = {}
    for category, proteins in protein_categories.items():
        console.rule(title=category, align="left")
        data = create_study_table(
            proteins=proteins,
            df=df_abundance,
            df_groups=df_groups,
            df_individuals=df_individuals,
        )
        df = pd.DataFrame(data)
        if drop_empty:
            # remove empty rows
            df.replace("", np.nan, inplace=True)
            df.dropna(inplace=True, axis="index", how="all")
            df.replace(np.nan, "", inplace=True)

        if sort_by_count:
            # sort proteins by count
            columns = df.columns
            counts = {col: (df[col] != "").sum() for col in columns}
            sorted_value_index = reversed(np.argsort(list(counts.values())))
            sorted_columns = [columns[i] for i in sorted_value_index]
            df = df[sorted_columns]

        # transpose table for reports
        df = df.T
        df.index = df.index.set_names("protein")
        df = df.reset_index()

        dfs_table[category] = df

    with pd.ExcelWriter(tables_xlsx) as writer:
        for category in protein_categories.keys():
            dfs_table[category].to_excel(writer, sheet_name=f"{category}", index=True)


def create_protein_tables(
    tables_xlsx: Path,
) -> None:
    """Statistics for all proteins."""
    # Load data

    df_abundance = pd.read_excel(DATA_XLSX, sheet_name="Abundance")
    del df_abundance["comments"]

    proteins = protein_info.get_proteins(df_abundance, uniprot=True)
    protein_categories = protein_info.get_protein_categories(proteins)

    dfs_table: Dict[str, pd.DataFrame] = {}
    for category, proteins in protein_categories.items():
        console.rule(title=category, align="left")

        all_info = []
        for protein_id in proteins:
            # protein_id = 'CYP1A2'
            # get all individual data for protein
            df_protein = df_abundance[df_abundance.protein == protein_id]
            # subset of abundance for individual protein
            df_protein = df_protein[~pd.isnull(df_protein.value)]

            data = df_protein["value"].values
            if data is not None and len(data) > 0:
                info = calculate_info_for_protein(
                    protein_id=protein_id, df_protein=df_protein
                )
                all_info.append(info)
            else:
                console.print(f"No data for protein_id: {protein_id}")

        df = pd.DataFrame(all_info)
        # sort by mean
        df.sort_values(by=["mean"], inplace=True, ascending=False)
        dfs_table[category] = df

    with pd.ExcelWriter(tables_xlsx) as writer:
        for category in protein_categories.keys():
            dfs_table[category].to_excel(writer, sheet_name=f"{category}", index=False)
    console.log(tables_xlsx)


def merge_tables(
    tables_studies_xlsx: Path, tables_proteins_xlsx: Path, tables_all_xlsx: Path
) -> None:
    """Merge tables."""
    # Load data
    dfs_studies = pd.read_excel(tables_studies_xlsx, sheet_name=None)
    dfs_proteins = pd.read_excel(tables_proteins_xlsx, sheet_name=None)

    with pd.ExcelWriter(tables_all_xlsx) as writer:
        for category in dfs_studies:
            # merge tables
            df1 = dfs_studies[category]
            console.log(df1)
            df2 = dfs_proteins[category]
            console.log(df2)
            df = df1.merge(df2, on="protein")
            console.log(df)
            # df = df.set_index("protein")

            df.to_excel(writer, sheet_name=f"{category}", index=False)

    console.log(tables_all_xlsx)


def calculate_info_for_protein(
    protein_id: str, df_protein: pd.DataFrame
) -> Dict[str, Any]:
    """Calculate statistics for given protein.

    Heterogeneity analysis similar to Archour2014
    """
    data: np.ndarray = df_protein["value"].values

    info = {
        "protein": protein_id,
        "mean": np.mean(data),
        "mode": stats.mode(data).mode,
        "cv": np.std(data) / np.mean(data),
        "wcv": "-",
        "min": data.min(),
        "max": data.max(),
        "n": len(data),
        "unit": "pmol/mg",
        "Q": "-",
        "I^2": "-",
        "df": "-",
        "heterogeneity": "-",
    }

    # degree of freedom
    study_ids = df_protein.study.unique()
    df = len(study_ids) - 1
    info["df"] = df

    # only calculate heterogeneity if degree of freedom > 0
    if df > 0:
        # weighted mean & weighted CV
        # console.print(df_protein)
        data_sets = []
        counts = []

        for study in study_ids:
            df_study = df_protein[df_protein.study == study]
            d = df_study["value"].values
            data_sets.append(d)
            counts.append(len(d))

        counts = np.array(counts)
        means = np.array([np.mean(d) for d in data_sets])
        cvs = np.array([np.std(d) / np.mean(d) for d in data_sets])

        # info["wmean"] = np.sum(counts*means)/np.sum(counts)
        info["wcv"] = np.sum(counts * cvs) / np.sum(counts)

        # Cochranes Q
        wjs = []
        for d in data_sets:
            if np.std(d) > 0.0:
                wj = 1 / (np.std(d) ** 2)
            else:
                # only single point, use mean/value as std: Coefficient of variant = 1 assumption
                wj = 1 / (np.mean(d) ** 2)
            wjs.append(wj)
        wjs = np.array(wjs)
        varwx = np.sum(wjs * means) / np.sum(wjs)
        Q = np.sum(wjs * (means - varwx) ** 2)
        info["Q"] = Q

        # heterogeneity
        # The degree of heterogeneity can be assessed using the I2 index (eq. 6)
        # proposed by Higgins and Thompson (2002). This index provides a percentage of
        # overall heterogeneity that can be interpreted as proposed by
        # Higgins et al. (2003) as follows:
        # - around 0%, no heterogeneity, : none
        # - around 25%, low heterogeneity, : low
        # - around 50%, moderate heterogeneity, : medium
        # - and around 75%, high heterogeneity. : high
        I2 = max(0, 100 * (Q - df) / Q)
        info["I^2"] = I2
        if 0 <= I2 < 25:
            heterogeneity = "none"
        elif 25 <= I2 < 50:
            heterogeneity = "low"
        elif 50 <= I2 < 75:
            heterogeneity = "medium"
        elif 75 <= I2:
            heterogeneity = "high"

        info["heterogeneity"] = heterogeneity

    # console.print(info)
    return info


def run_all_tables() -> None:
    """Create all tables."""
    tables_studies_xlsx: Path = RESULTS_DIR / "table_studies.xlsx"
    tables_proteins_xlsx: Path = RESULTS_DIR / "table_proteins.xlsx"
    tables_all_xlsx: Path = RESULTS_DIR / "table_all.xlsx"

    create_study_tables(tables_xlsx=tables_studies_xlsx)
    create_protein_tables(tables_xlsx=tables_proteins_xlsx)

    merge_tables(
        tables_studies_xlsx,
        tables_proteins_xlsx,
        tables_all_xlsx=tables_all_xlsx,
    )


if __name__ == "__main__":
    run_all_tables()



"""Module for logging.

Using rich for output formating.
"""
import logging

from rich.logging import RichHandler

from protein_distribution.console import console


def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:
    """Get new custom logger for name."""
    formatter = logging.Formatter(
        fmt="%(message)s",
        datefmt="[%X]",
    )

    # handler = logging.StreamHandler()
    handler = RichHandler(
        markup=False, rich_tracebacks=True, show_time=False, console=console
    )
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)
    return logger



"""Process all data and files for release."""
from protein_distribution.dataio import run_all_dataio
from protein_distribution.tables import run_all_tables
from protein_distribution.uniprot import query_metadata


if __name__ == "__main__":
    query_metadata()
    run_all_dataio()
    run_all_tables()
